/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ASSEMBLER_CPP
#define ASSEMBLER_CPP

#include "Assembler.h"


#include "Read.h"
#include "ScoredSeqSubseq.h"
#include "ScoredSeqNested.h"
#include "ScoredSeqMonoScore.h"
#include <sstream>
#include <omp.h>

ParamsMapping* Assembler::getDefaultMapParams(){ return new ParamsMapping(25000, 0.9); }
AlignmentScoreMatrix* Assembler::getDefaultScoreMatrix(){ return new AlignmentScoreMatrix(1,-2,-5,-2); }
ParamsMinOverlap* Assembler::getDefaultParamsMinOverlap(){ return new ParamsMinOverlap(35,20,5); }


Assembler::Assembler(){}
Assembler::Assembler(set<ParameterizedReadFile*>* pairedReadFiles,
		     set<ParameterizedInitialFile*>* initialContigFiles,
		     ParamsCycleManagement* cycleInfo, OutputFile* outfile,
		     AssemblerListener* listener, int maxThreadsPerFile,
		     ParamsDeBruijn* paramsDeBruijn,
		     ReadPairFilter* rpf) :
  _listener(listener),
  _maxThreadsPerFile(maxThreadsPerFile),
  _outfile(outfile) {
  constructorHelper(pairedReadFiles,initialContigFiles);
  ParamsMinOverlap* defPmo = getDefaultParamsMinOverlap();
  ParamsMinFractId* defPmf = ParamsMinFractId::getDefault();
  AlignmentScoreMatrix* defAsm = getDefaultScoreMatrix();
  _miniAlignParams = new ParamsAlignment(defPmo, defPmf, defAsm);
  _metaAlignParams = new ParamsAlignment(defPmo, defPmf, defAsm);
  _cycleInfo = new ParamsCycleManagement(cycleInfo);
  _paramsDeBruijn = new ParamsDeBruijn(paramsDeBruijn);
  delete defPmo;
  delete defPmf;
  delete defAsm;
  _rpf = rpf->copy();
}
Assembler::Assembler(set<ParameterizedReadFile*>* pairedReadFiles,
		     set<ParameterizedInitialFile*>* initialContigFiles,
		     ParamsCycleManagement* cycleInfo, OutputFile* outfile,
		     AssemblerListener* listener, int maxThreadsPerFile,
		     ParamsAlignment* miniAlignParams,
		     ParamsDeBruijn* paramsDeBruijn,
		     ReadPairFilter* rpf) :
  _listener(listener),
  _maxThreadsPerFile(maxThreadsPerFile),
  _outfile(outfile) {
  constructorHelper(pairedReadFiles,initialContigFiles);
  ParamsMinOverlap* defPmo = getDefaultParamsMinOverlap();
  ParamsMinFractId* defPmf = ParamsMinFractId::getDefault();
  AlignmentScoreMatrix* defAsm = getDefaultScoreMatrix();
  _miniAlignParams = new ParamsAlignment(miniAlignParams);
  _metaAlignParams = new ParamsAlignment(defPmo, defPmf, defAsm);
  _cycleInfo = new ParamsCycleManagement(cycleInfo);
  _paramsDeBruijn = new ParamsDeBruijn(paramsDeBruijn);
  delete defPmo;
  delete defPmf;
  delete defAsm;
  _rpf = rpf->copy();
}
Assembler::Assembler(set<ParameterizedReadFile*>* pairedReadFiles,
		     set<ParameterizedInitialFile*>* initialContigFiles,
		     ParamsCycleManagement* cycleInfo, OutputFile* outfile,
		     AssemblerListener* listener, int maxThreadsPerFile,
		     ParamsAlignment* miniAlignParams,
		     ParamsAlignment* metaAlignParams,
		     ParamsDeBruijn* paramsDeBruijn,
		     ReadPairFilter* rpf) :
  _listener(listener),
  _maxThreadsPerFile(maxThreadsPerFile),
  _outfile(outfile) {
  constructorHelper(pairedReadFiles,initialContigFiles);
  _miniAlignParams = new ParamsAlignment(miniAlignParams);
  _metaAlignParams = new ParamsAlignment(metaAlignParams);
  _cycleInfo = new ParamsCycleManagement(cycleInfo);
  _paramsDeBruijn = new ParamsDeBruijn(paramsDeBruijn);
  _rpf = rpf->copy();
}

void Assembler::constructorHelper(set<ParameterizedReadFile*>* pairedReadFiles,
				  set<ParameterizedInitialFile*>* initialContigFiles){
  ParameterizedReadFile* rfArray[ pairedReadFiles->size() ];
  int prfCount = 0;
  for (set<ParameterizedReadFile*>::iterator it = pairedReadFiles->begin();
       it != pairedReadFiles->end(); ++it){
    rfArray[prfCount] = *it;
    prfCount++;
  }
  long rfSizes[ prfCount ];
  #pragma omp parallel for schedule(dynamic)
  for (int n = 0; n < prfCount; ++n){ rfSizes[n] = rfArray[n]->numReads(); }
  int numThreads = omp_get_max_threads();
  long readTotal = 0;
  for (int n = 0; n < prfCount; ++n){ readTotal += rfSizes[n]; }
  long threadDenom = readTotal / numThreads;
  if (threadDenom == 0){ threadDenom = 1; }
  #pragma omp parallel for schedule(dynamic)
  for (int n = 0; n < prfCount; ++n){
    int rfThreads = rfSizes[n] / threadDenom;
    if (rfThreads == 0){ rfThreads = 1; }
    else if (rfThreads > _maxThreadsPerFile){ rfThreads = _maxThreadsPerFile; }
    set<ParameterizedReadFile*> splitFiles;
    rfArray[n]->splitFile(&splitFiles, rfThreads);
    #pragma omp critical (Assembler_addRfsToSet)
    { _pairedReadFiles.insert( splitFiles.begin(), splitFiles.end() ); }
  }

  //_pairedReadFiles.insert( pairedReadFiles->begin(), pairedReadFiles->end() );
  _initialContigFiles.insert( initialContigFiles->begin(), initialContigFiles->end() );
}
Assembler::~Assembler(){
  for (set<ScoredSeq*>::iterator it = _currentCycleContigs.begin(); it != _currentCycleContigs.end(); ++it){ (*it)->deepDelete(); }
  delete _miniAlignParams;
  delete _metaAlignParams;
  delete _cycleInfo;
  delete _paramsDeBruijn;
  for (set<ParameterizedReadFile*>::iterator it = _pairedReadFiles.begin();
       it != _pairedReadFiles.end(); ++it){ delete (*it); }
  delete _rpf;
}



void Assembler::runAssembly(){
  _listener->beginAssembly("BEGIN ASSEMBLY");
  for (int n = 0; n < _cycleInfo->totalCycles(); ++n){
    _listener->beginCycle(n+1);
    generateInitialContigs(n);
    extendOneCycle(n);

    // make the number into an addendum string and write the file
    if ( _cycleInfo->outputThisCycle(n+1) ){
      stringstream addendum;
      addendum << "cycle" << n+1;
      writeOutfile( addendum.str() );
    }
    _listener->endCycle(&_currentCycleContigs);
  }
  // delete the contigs from the prior cycle in the end
  for (set<ScoredSeq*>::iterator it = _priorCycleContigs.begin(); it != _priorCycleContigs.end(); ++it){
    (*it)->deepDelete();
  }
  _listener->endAssembly("END ASSEMBLY");
}

void Assembler::writeOutfile(string addendum){
  if (_outfile->getName(addendum) != ""){
    _listener->writeOutfile( _outfile->getName(addendum) );
  }
  _outfile->open(addendum);
  _outfile->writeContigs( &_currentCycleContigs, &_currentCycleLeftovers );
  _outfile->close();
}



void Assembler::getContigs(set<ScoredSeq*>* outputContigs){
  for (set<ScoredSeq*>::iterator it = _currentCycleContigs.begin(); it != _currentCycleContigs.end(); ++it){
    outputContigs->insert( (*it)->shallowCopy() );
  }
}



void Assembler::findLeftoverContigs(set<ScoredSeq*>* current, set<ScoredSeq*>* prior, set<ScoredSeq*>* leftover){
  // collect the sequences of the contigs being input into this assembly round
  set<string> seqsFromPrior;
  for (set<ScoredSeq*>::iterator it = prior->begin(); it != prior->end(); ++it){
    seqsFromPrior.insert( (*it)->getSeq('+') );
  }
  // now identify the contigs from the current cycle that didn't make progress from the last
  for (set<ScoredSeq*>::iterator it = current->begin(); it != current->end(); ++it){
    if (seqsFromPrior.count( (*it)->getSeq('+') ) > 0 or 
	seqsFromPrior.count( (*it)->getSeq('-') ) > 0 ){
      leftover->insert( (*it) );
    }
  }
}


void Assembler::generateInitialContigs(int cycleNum){
  set<ScoredSeq*> initialSeqs;

  // put the current contigs in with the new initial seqs
  initialSeqs.insert( _currentCycleContigs.begin(), _currentCycleContigs.end() );
  _currentCycleContigs.clear();

  long initialCount = initialSeqs.size();
  for (set<ParameterizedInitialFile*>::iterator rfIt = _initialContigFiles.begin(); rfIt != _initialContigFiles.end(); ++rfIt){
    (*rfIt)->getContigs(&initialSeqs, cycleNum);
  }

  // TRIM
  _cycleInfo->executeTrim( &initialSeqs, -1 );

  if ( initialCount == initialSeqs.size() ){
    _currentCycleContigs.insert( initialSeqs.begin(), initialSeqs.end() );
  } else {
    _listener->beginAddInitialContigs(&initialSeqs);

    // condense (assemble) the initial contigs (assembly is double-stranded)
    AssemblyJobFactory * contigFactory = new AssemblyJobFactory(_metaAlignParams,_paramsDeBruijn,AssemblyJob::doubleStranded,_listener);
    AssemblyJob * ajr2 = contigFactory->redundancyJob( &initialSeqs );
    ajr2->runJob(AssemblyJob::threaded);
    // add the output of the initial assembly to the collection
    // currentCycleContigs is the initial input; priorCycleContigs is initially empty
    ajr2->allAssembledSeqs( &_currentCycleContigs );

    set<ScoredSeq*> usedSeqs;
    ajr2->discardedInputSeqs( &usedSeqs );
    for (set<ScoredSeq*>::iterator it = usedSeqs.begin(); it != usedSeqs.end(); ++it){ delete (*it); }
    delete ajr2;

    _listener->endAddInitialContigs(&_currentCycleContigs);
    delete contigFactory;
  }
}


void Assembler::extendOneCycle(int cycleNum){

  // identify the contigs that have not changed since the previous cycle
  set<ScoredSeq*> leftoverContigs;
  findLeftoverContigs(&_currentCycleContigs, &_priorCycleContigs, &leftoverContigs);
  for (set<ScoredSeq*>::iterator it = _priorCycleContigs.begin(); it != _priorCycleContigs.end(); ++it){ (*it)->deepDelete(); }
  _priorCycleContigs.clear();

  ostringstream outMessage;
  outMessage << "num contigs leftover: " << leftoverContigs.size();
  _listener->message(outMessage.str());

  // isolate just the read files that are appropriate for this cycle
  set<ParameterizedReadFile*> thisCycleReadFiles;
  for (set<ParameterizedReadFile*>::iterator it = _pairedReadFiles.begin(); it != _pairedReadFiles.end(); ++it){
    if ( (*it)->useThisCycle(cycleNum) ){ thisCycleReadFiles.insert( (*it) ); }
  }

  // this will be augmented later
  ReadPairFilter* rpf = _rpf->copy();

  // create the extend cycle and copy over the contig sets
  ExtendCycle* ec = new ExtendCycle(&thisCycleReadFiles, rpf, &_currentCycleContigs, &leftoverContigs, _listener);
  delete rpf;
  _priorCycleContigs.insert(_currentCycleContigs.begin(), _currentCycleContigs.end());
  _currentCycleContigs.clear();

  // run the cycle and collect, filter, and trim the output
  EcoFilter* sizeFilter = _cycleInfo->getSizeFilter();
  ec->runCycle(_miniAlignParams, _metaAlignParams, _paramsDeBruijn, _cycleInfo->maxContigLinkages(), sizeFilter);
  delete sizeFilter;
  ec->getContigs(&_currentCycleContigs);
  delete ec;

  filterOutputContigs(cycleNum);
  _cycleInfo->executeTrim( &_currentCycleContigs, cycleNum );

  // identify those contigs that did not change during this cycle
  _currentCycleLeftovers.clear();
  findLeftoverContigs(&_currentCycleContigs, &_priorCycleContigs, &_currentCycleLeftovers);
}


void Assembler::filterOutputContigs(int cycleNum){
  // evaluate contigs based on linkage scores
  set<ScoredSeq*> passingContigs;
  set<ScoredSeq*> failingContigs;
  EcoFilter* fullFilter = _cycleInfo->getFullEcoFilter();
  fullFilter->filterSeqs(&_currentCycleContigs, &passingContigs, &failingContigs, cycleNum);
  delete fullFilter;
  for (set<ScoredSeq*>::iterator seqIt = failingContigs.begin(); seqIt != failingContigs.end(); ++seqIt){ (*seqIt)->deepDelete(); }
  _currentCycleContigs.clear();
  _currentCycleContigs.insert( passingContigs.begin(), passingContigs.end() );
}

long Assembler::getTotalSize(){
  long count = 0;
  for (set<ScoredSeq*>::iterator itS = _currentCycleContigs.begin(); itS != _currentCycleContigs.end(); ++itS){
    count += (*itS)->size();
  }
  return count;
}

long Assembler::getMaxSize(){
  long maxSize = 0;
  for (set<ScoredSeq*>::iterator itS = _currentCycleContigs.begin(); itS != _currentCycleContigs.end(); ++itS){
    if ( (*itS)->size() > maxSize ){ maxSize = (*itS)->size(); }
  }
  return maxSize;
}


long Assembler::getN50(){
  vector<long> contigLengths;
  for (set<ScoredSeq*>::iterator seqIt = _currentCycleContigs.begin(); seqIt != _currentCycleContigs.end(); ++seqIt){
    contigLengths.push_back( (*seqIt)->size() );
  }
  sort( contigLengths.begin(), contigLengths.end() );
  long halfSize = getTotalSize() / 2;

  long n50 = 0;
  vector<long>::iterator sizeIt = contigLengths.begin();
  while ( halfSize > 0 and sizeIt != contigLengths.end() ){
    n50 = (*sizeIt); // current n50; will only be updated if half the total size is not reached
    halfSize -= (*sizeIt);
    ++sizeIt;
  }
  return n50;
}
bool reverseN50Helper(long i,long j) { return (i>j); }


#endif



/*
THE AUTHOR: generated by www.degraeve.com/img2txt.php
::::;::::;::::i:,,,tiiiitt;;,;,,,;,,,,,,
::::;::::;:,i:i;i;;jfGLfji:,,;,,,;,,,,,,
::::;::::,:;,jGGfLGLLGLffLtit;,,,;,,,,,,
::::,::::;;jLGGGEEEDGGGLjfLLt;::,;,,,,,,
::::,::::ijLDGDGLKWKEDGGLffLt;,::;:,,,,,
::::,:::ifDGDDGGDEKWKEDLLLGGft:::;::,,,:
::::,::;tiLDDEDDDDDEKKKKEDLLLLi,:;,::,,,
:..:,::jitEDKKDGLLLLGGDEEDDfGLf,:;::,,,,
:::::::L,LDEWDLLffffLLGDDDDfLGLi:,:::,,:
..::::;jiDDKKLLLfffffffLGDEGLLLf:,:.::::
...:,.;:tfEWDGLLffffjjjjjLEDLDLf::...:.:
:...:..;iEEKGGLLfffjfjtttjLDDDLf::...::.
....:t:.LEEKLGLLfffjjjttiitDEEGLL:....:.
:...:LGLEDDDGLLffLLffjttiiijEKGj;:......
:...:jKKKKWELDDDKEDLfjjjjjiiLEEj::......
..:.tiEEKKWDWWKKWKKDffDKDGj;DEDG;:....:.
..:;iGKEKKKLWKEKKWKELEWEEGG#EEDEKtj;,;;,
ii;iiLEEDKDGKDEEEKEDjEEEGfGt#EDDEDi,,,,,
i;;iiiKKKKDLGGGDEEELtWLEEGjtGGGDDL,,,,,,
iiiiiiGfGEDLLLLGDDELiGfLLftfLGEDLfi;;iii
iiiiiiitKDDLLGEEEEGfiiGLLji;LGEGijiiiiii
iiiitiitLfDLGGDDDLGLiitELfi;jDGiijiiiiii
iiiiiiiittGLGDEEEEEDfLfDDLtiGLiiijiiiitt
tttttiiiitEGGGEDDEDDGffLGDtiGLi;itiiiiii
fffffffLGfEDGGDDWEEGGLLKGLjLjfjiitiiiiii
ttttttttitLEGGLGDEGLjDftftiDjiiiitiiiiii
#########WL#GGGGGDDGGjitftDKjiiiitiiiitt
WWWWWWWWKKfWLGGGGGGGLjtttftiiiiiijtttttt
WWWWWDWWWfEKLGGDLGLLfjtittiiitiiijtttttt
WWWWWKKKfGEDGGGDDGLLfttjitiiitiiitiiiitt
WWWWWKLfLfLDGGGDEDGGLjfjjiittttttftttttj
WWWLLGGLLLGEGGGGDDDDDLjtKG;;tttttftttjjj
jjfjjLLfLDKKGGGGGDGGjtttWD;jt;tttftjjjjj
ttjLLLffLLDKDGDGGGGfjtttKf;Lfi,;tjjjjjjj
tjDLDfGLtjGEDGDGDDDfjjjGWL;;tL;,;:tjjjjf
fGLDLLtjjtfDDDDDDDGLjfLEED,iGfLt,,;::;tj
jfLDfjGGGLtLEDGDDGGLLGELfGt;ijGD;ti;;,,j
LjfDfLGKWWWjfDGGDDDDGDjGjft;;fDLij,ti,:D
fDLLLDEKKWKGLjGLGGDGGtLGjfD:itfjfjt,j;tD
LLDGGEDWDEDGDjjtijLitfDjfjKKitDDjff,fiiL
DjLDfDGGGGEGGGjjttjtGfffGDKWKfGDiGL;tiji
jDLjDfGLLLGDGEftiiijtfLLfGEEWttLfLLf;jit
fLEGLjGjffGGGGEftitjfjGjtjjLKiLGttGftjit
GjfDjffjjfLLLLjDfjfffGLti;jjtiGL;fELfijj
LfjGfffjjfLLfffjDLLfGGjji;ttfLLLtEDLLjGj
fLGGffjjjjfGjjjjfGjLGLttiijtttjGjjfDiLLi
LffLjfjjjffGjjjfLGfGDGjtttttjtjjtfLLfff;
LfffjjjjfffGjjffLtfEDDjtfjjtjjffjffjtfjt
GLffjjfjfffGjjfftiGDDGLfffjtjjfLjfGGjjLj
*/
