/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This is the programmatic interface.
 */


#ifndef ASSEMBLER_H
#define ASSEMBLER_H

#include <set>
#include <algorithm>
#include <vector>
#include "ScoredSeq.h"
#include "ReadFile.h"
#include "ParamsMapping.h"
#include "ParamsCycleManagement.h"
#include "ParamsAlignment.h"
#include "AssemblyJob.h"
#include "AssemblyJobFactory.h"
#include "ExtendCycle.h"
#include "OutputFile.h"
#include "ParameterizedReadFile.h"
#include "ParameterizedInitialFile.h"
#include "EcoFilter.h"
#include "AssemblerListener.h"
#include "ParamsDeBruijn.h"
#include "ReadPairFilter.h"
using namespace::std;

class Assembler {

 public:
  Assembler();
  Assembler(set<ParameterizedReadFile*>* readFileSet,
	    set<ParameterizedInitialFile*>* initialContigFiles,
	    ParamsCycleManagement* cycleInfo, OutputFile* outfile,
	    AssemblerListener* listener, int maxThreadsPerFile,
	    ParamsDeBruijn* paramsDeBruijn,
	    ReadPairFilter* rpf);
  Assembler(set<ParameterizedReadFile*>* pairedReadFiles,
	    set<ParameterizedInitialFile*>* initialContigFiles,
	    ParamsCycleManagement* cycleInfo, OutputFile* outfile,
	    AssemblerListener* listener, int maxThreadsPerFile,
	    ParamsAlignment* miniAlignParams,
	    ParamsDeBruijn* paramsDeBruijn,
	    ReadPairFilter* rpf);
  Assembler(set<ParameterizedReadFile*>* pairedReadFiles,
	    set<ParameterizedInitialFile*>* initialContigFiles,
	    ParamsCycleManagement* cycleInfo, OutputFile* outfile,
	    AssemblerListener* listener, int maxThreadsPerFile,
	    ParamsAlignment* miniAlignParams,
	    ParamsAlignment* metaAlignParams,
	    ParamsDeBruijn* paramsDeBruijn,
	    ReadPairFilter* rpf);
  ~Assembler();

  void runAssembly();
  void getContigs(set<ScoredSeq*>* outputContigs);

  long getTotalSize();
  long getMaxSize();
  long getN50();

  static ParamsMapping* getDefaultMapParams();
  static AlignmentScoreMatrix* getDefaultScoreMatrix();
  static ParamsMinOverlap* getDefaultParamsMinOverlap();
  //static ParamsMinFractId* getDefaultParamsMinFractId();

 private:
  void constructorHelper(set<ParameterizedReadFile*>* pairedReadFiles,
			 set<ParameterizedInitialFile*>* initialContigFiles);

  // If numInitialContigs==0, then all of the contigs in all the files will be used.
  // Same goes for if numInitialContigs > the num of sequences in the files.
  void generateInitialContigs(int cycleNum);
  void extendOneCycle(int cycleNum);
  void writeOutfile(string addendum);
  void filterOutputContigs(int cycleNum);
  bool reverseN50Helper(long i,long j);
  // helps sort sequences - only modifies leftover by adding components of current
  void findLeftoverContigs(set<ScoredSeq*>* current, set<ScoredSeq*>* prior, set<ScoredSeq*>* leftover);

  set<ParameterizedReadFile*> _pairedReadFiles;
  set<ParameterizedInitialFile*> _initialContigFiles;

  set<ScoredSeq*> _priorCycleContigs;
  set<ScoredSeq*> _currentCycleContigs;
  set<ScoredSeq*> _priorCycleLeftovers;
  set<ScoredSeq*> _currentCycleLeftovers;

  ParamsAlignment* _miniAlignParams;
  ParamsAlignment* _metaAlignParams;
  ParamsDeBruijn* _paramsDeBruijn;
  ParamsCycleManagement* _cycleInfo;
  ReadPairFilter* _rpf;

  OutputFile* _outfile;
  AssemblerListener* _listener;

  int _maxThreadsPerFile;
};

#endif
