/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ASSEMBLYJOBNEARPERFECT_CPP
#define ASSEMBLYJOBNEARPERFECT_CPP


#include "AssemblyJobNearPerfect.h"

#include "AssemblyException.h"
#include <queue>
#include <string>
#include <iostream>
#include <math.h>
using namespace::std;

AssemblyJobNearPerfect::AssemblyJobNearPerfect(){}
AssemblyJobNearPerfect::AssemblyJobNearPerfect(set<ScoredSeq*>* seqs,
					       float minFractId,
					       AssemblyJob::AssemblyStrandedness strandedness) :
  _wasRun(false),
  _inputRemoved(false),
  _novelRemoved(false),
  _strandedness(strandedness),
  _maxFractMis(1.0 - minFractId){
  _inputSeqs.insert(seqs->begin(),seqs->end());
  // true if mismatches <= maxMisNum
  _mismatchCountAsm = new AlignmentScoreMatrix(0,1,0,0); // mismatch counts as 1
  _matchCountAsm = new AlignmentScoreMatrix(1,0,0,0); // mismatch counts as 1
}


AssemblyJobNearPerfect::~AssemblyJobNearPerfect(){
  delete _mismatchCountAsm;
  delete _matchCountAsm;
}


void AssemblyJobNearPerfect::makeShallow(bool shallowDelete){
  if (_wasRun){
    throw AssemblyException::CallingError("AJNP::makeShallow can't be called after the job was run.");
  }
  set<ScoredSeq*> shallowCopies;
  // buffer all input, then copy
  for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
    (*it)->bottomBuffer();
  }
  for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
    shallowCopies.insert( (*it)->shallowCopy() );
    (*it)->deepUnbuffer();
  }
  if ( shallowDelete){
    for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
      delete (*it);
    }
  }
  // replace the input contents with the shallow copies
  _inputSeqs.clear();
  _inputSeqs.insert( shallowCopies.begin(), shallowCopies.end() );
}


bool AssemblyJobNearPerfect::wasRun(){ return _wasRun; }

void AssemblyJobNearPerfect::allInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
}

void AssemblyJobNearPerfect::remainingInputSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobNP::remainingInputSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobNP::remainingInputSeqs; seqs were already removed.");
  }
  outsideSet->insert( _retainedSeqs.begin(), _retainedSeqs.end() );
}

void AssemblyJobNearPerfect::discardedInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _usedSeqs.begin(), _usedSeqs.end() );
}

void AssemblyJobNearPerfect::newlyAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobNP::newlyAssembledSeqs; job hasn't run yet.");
  }
  if ( _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobNP::newlyAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}

void AssemblyJobNearPerfect::allAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobNP::allAssembledSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved or _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobNP::allAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _retainedSeqs.begin(), _retainedSeqs.end() );
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}



void AssemblyJobNearPerfect::OK(){
}







void AssemblyJobNearPerfect::runJob(AssemblyJob::AssemblyThreadedness threadedness){
  if (! _wasRun){

    // STEP 1: organize input data by sequence length
    map<long, set<ScoredSeq*>* > lenToSeqSet;
    for (set<ScoredSeq*>::iterator inputIt = _inputSeqs.begin(); inputIt != _inputSeqs.end(); ++inputIt){
      ScoredSeq* inSeq = (*inputIt);
      map<long, set<ScoredSeq*>* >::iterator lenToSeqsIt = lenToSeqSet.find( inSeq->size() );
      if ( lenToSeqsIt == lenToSeqSet.end() ){
	set<ScoredSeq*>* newEntry = new set<ScoredSeq*>;
	newEntry->insert( inSeq );
	lenToSeqSet.insert( pair<int, set<ScoredSeq*>* > (inSeq->size(), newEntry) );
      } else {
	lenToSeqsIt->second->insert( inSeq );
      }
    }

    // STEP 2: run each set through a subroutine that consolidates its contents based on near-matches

    // organize into arrays to facilitate threading
    long indexToSeqLen[ lenToSeqSet.size() + 1 ];
    set<ScoredSeq*>** indexToSeqSet = new set<ScoredSeq*>*[ lenToSeqSet.size() + 1 ];
    long numLengthSets = 0;
    for (map<long, set<ScoredSeq*>* >::iterator lenToSeqsIt = lenToSeqSet.begin(); lenToSeqsIt != lenToSeqSet.end(); ++lenToSeqsIt){
      indexToSeqLen[numLengthSets] = lenToSeqsIt->first;
      indexToSeqSet[numLengthSets] = lenToSeqsIt->second;
      numLengthSets++;
    }

    // run the consolidation threaded if appropriate
    if (threadedness == AssemblyJob::threaded){
      #pragma omp parallel for schedule(dynamic)
      for (long n = 0; n < numLengthSets; ++n){
	subAssembly( indexToSeqLen[n], indexToSeqSet[n] );
      }
    } else {
      for (long n = 0; n < numLengthSets; ++n){
	subAssembly( indexToSeqLen[n], indexToSeqSet[n] );
      }
    }

    // STEP 3: clean up
    // re-consolidate and sort the sequences
    for (long n = 0; n < numLengthSets; ++n){
      for (set<ScoredSeq*>::iterator it = indexToSeqSet[n]->begin(); it != indexToSeqSet[n]->end(); ++it){
	if ( _inputSeqs.count(*it) == 0 ){ _novelSeqs.insert(*it); }
	else { _retainedSeqs.insert(*it); }
      }
      delete indexToSeqSet[n];
    }
    delete [] indexToSeqSet;
    // figure out which of the input seqs was used
    for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
      if (_retainedSeqs.count(*it) == 0){ _usedSeqs.insert(*it); }
    }

    _wasRun = true;
  }
  //OK();
}


AssemblyJobNearPerfect::WindowStartAndLenCarrier::WindowStartAndLenCarrier(){}
AssemblyJobNearPerfect::WindowStartAndLenCarrier::WindowStartAndLenCarrier(long seqLen, long maxMisNum){
  // calculate the number of windows
  // scale the number of windows with seqLen (also considers best-case scenario with maxMisNum)
  long maxNumWindows = maxMisNum + 1;
  long baseCase = 50;
  long minNumWindows;
  if (seqLen < baseCase){ minNumWindows = 2; }
  else{ minNumWindows = long( log10( double(seqLen) / baseCase ) / log10( 2 ) ) + 2; }
  if (maxNumWindows < minNumWindows){ _numWindows = maxNumWindows; }
  else { _numWindows = minNumWindows; }

  // find the window dimensions
  _winStart = new long[_numWindows];
  _winLen = new long[_numWindows];
  // delineate the windows as evenly as possible
  _winStart[0] = 0;
  for (int n = 1; n < _numWindows; ++n){
    _winStart[n] = n * seqLen / _numWindows;
    _winLen[n-1] = _winStart[n] - _winStart[n-1];
  }
  _winLen[_numWindows-1] = seqLen - _winStart[_numWindows-1];
}
AssemblyJobNearPerfect::WindowStartAndLenCarrier::~WindowStartAndLenCarrier(){
  delete [] _winStart;
  delete [] _winLen;
}





void AssemblyJobNearPerfect::subAssembly(long seqLen, set<ScoredSeq*>* monoLenSeqSet){
  ScoredSeqNormalized** withWrappers = new ScoredSeqNormalized*[ monoLenSeqSet->size() + 1 ];
  long ssnIndex = 0;
  for (set<ScoredSeq*>::iterator it = monoLenSeqSet->begin(); it != monoLenSeqSet->end(); ++it){
    withWrappers[ssnIndex] = new ScoredSeqNormalized( *it );
    ssnIndex++;
  }
  withWrappers[ssnIndex] = NULL;

  subAssembly( seqLen, withWrappers );

  monoLenSeqSet->clear();
  ssnIndex = 0;
  while ( withWrappers[ssnIndex] != NULL ){
    monoLenSeqSet->insert( withWrappers[ssnIndex]->getNested() );
    delete withWrappers[ssnIndex];
    ssnIndex++;
  }
  delete [] withWrappers;
}


// MODIFIES: monoLenSeqSet
void AssemblyJobNearPerfect::subAssembly(long seqLen, ScoredSeqNormalized** monoLenSeqInput){
  long maxMisNum = long(float(seqLen) * _maxFractMis);


  // local so that it can be emptied
  set<ScoredSeqNormalized*> monoLenSeqSet;
  long ssnCount = 0;
  while ( monoLenSeqInput[ssnCount] != NULL ){
    monoLenSeqSet.insert( monoLenSeqInput[ssnCount] );
    ssnCount++;
  }
  
  WindowStartAndLenCarrier* wslc = new WindowStartAndLenCarrier(seqLen, maxMisNum);
  map<string, set<ScoredSeqNormalized*>* >* windowToSubstrToSeqs = new map<string, set<ScoredSeqNormalized*>* >[ wslc->_numWindows ];

  for (long ssnIndex = 0; ssnIndex < ssnCount; ++ssnIndex){
    ScoredSeqNormalized* query = monoLenSeqInput[ssnIndex];
    monoLenSeqSet.erase( query );

    // get all the legit matches for each strand of the query
    map<ScoredSeqNormalized*,Alignment*> legitMatches; // key is seqB: seqA is the query seq; seqB is the found match
    // NOTE: this may NOT contain both (+) and (-) orientation matches to a given seqB

    if (_strandedness == AssemblyJob::doubleStranded){
      getLegitMatchesHelper( query, '+', wslc, maxMisNum, windowToSubstrToSeqs, &legitMatches );
      getLegitMatchesHelper( query, '-', wslc, maxMisNum, windowToSubstrToSeqs, &legitMatches );
    } else {
      getLegitMatchesHelper( query, '+', wslc, maxMisNum, windowToSubstrToSeqs, &legitMatches );
    }

    // define the set of sequences to be removed and add the new sequences to the _novel set
    // NOTE: the query is not yet in the search structure
    set<ScoredSeqNormalized*> seqsToRemove;
    // either the query or the newly-consolidated sequence will be added
    ScoredSeqNormalized* consolidatedSeq;
    if ( legitMatches.size() > 0 ){
      set<ScoredSeq*> seqsToInsert;
      for (map<ScoredSeqNormalized*,Alignment*>::iterator legitIt = legitMatches.begin(); legitIt != legitMatches.end(); ++legitIt){
	seqsToRemove.insert( legitIt->first );
	ScoredSeq* newContig =  AssemblyJobHelperStrategies::combine( legitIt->second, legitMatches.size() );
	seqsToInsert.insert( newContig );
	delete legitIt->second;
      }
      consolidatedSeq = new ScoredSeqNormalized( consolidateSeqsHelper(&seqsToInsert, seqLen) );
      consolidatedSeq->buffer();
      ScoredSeq* innerQuery = query->getNested();
      delete query;
      if (_inputSeqs.count(innerQuery) == 0){ delete innerQuery; }
    } else {
      consolidatedSeq = query;
    }

    // insert the new seq (there will be only one by definition)
    monoLenSeqSet.insert( consolidatedSeq );
    insertSeqHelper( consolidatedSeq, wslc, windowToSubstrToSeqs );

    // delete the old seqs if that is allowed
    for (set<ScoredSeqNormalized*>::iterator oldSeqIt = seqsToRemove.begin(); oldSeqIt != seqsToRemove.end(); ++oldSeqIt){
      monoLenSeqSet.erase( *oldSeqIt );
      removeSeqHelper( (*oldSeqIt), wslc, windowToSubstrToSeqs );
    }
  }

  for (long n = 0; n < wslc->_numWindows; ++n){
    for (map<string, set<ScoredSeqNormalized*>* >::iterator it = windowToSubstrToSeqs[n].begin(); it != windowToSubstrToSeqs[n].end(); ++it){
      delete it->second;
    }
  }
  delete wslc;
  delete [] windowToSubstrToSeqs;

  // modify the input set to just have the now legit contig seqs
  {
    long ssnCount = 0;
    for (set<ScoredSeqNormalized*>::iterator it = monoLenSeqSet.begin(); it != monoLenSeqSet.end(); ++it){
      monoLenSeqInput[ssnCount] = *it;
      ssnCount++;
    }
    monoLenSeqInput[ssnCount] = NULL;
  }
}




// REQUIRES: all input seqs are in the sense in which they will be combined
ScoredSeq* AssemblyJobNearPerfect::consolidateSeqsHelper(set<ScoredSeq*>* localSet, long seqLen){
  if (localSet->size() == 0){
    throw AssemblyException::ArgError("cannot consolidate an empty set of sequences");
  }
  set<ScoredSeq*>::iterator seqIt = localSet->begin();
  ScoredSeq* resultSeq = (*seqIt); // this will end up embodying the full consolidation
  ++seqIt;
  while ( seqIt != localSet->end() ){
    Alignment* al = new Alignment( resultSeq, (*seqIt), '+' );
    al->addLinkageBlock(resultSeq,0,0,seqLen);
    al->lock();
    ScoredSeq* newContig =  AssemblyJobHelperStrategies::combine( al );
    // this will ensure that the input sequences are not deleted
    delete al->seqA();
    delete al->seqB();
    delete al;
    resultSeq = newContig;
    ++seqIt;
  }
  localSet->clear();
  return resultSeq;
}




// add the combined seq to the front/back special map collections
void AssemblyJobNearPerfect::insertSeqHelper(ScoredSeqNormalized* newContig,
					     WindowStartAndLenCarrier* wslc,
					     map<string, set<ScoredSeqNormalized*>* >* windowToSubstrToSeqs){
  string newString = newContig->getSeq('+');
  for (int n = 0; n < wslc->_numWindows; ++n){
    string subSeq = newString.substr(wslc->_winStart[n],wslc->_winLen[n]);
    map<string,set<ScoredSeqNormalized*>* >::iterator matchIt = windowToSubstrToSeqs[n].find(subSeq);
    if ( matchIt == windowToSubstrToSeqs[n].end() ){
      set<ScoredSeqNormalized*>* newSeqSet = new set<ScoredSeqNormalized*>;
      newSeqSet->insert( newContig );
      windowToSubstrToSeqs[n].insert( pair<string, set<ScoredSeqNormalized*>* > (subSeq, newSeqSet) );
    } else {
      matchIt->second->insert( newContig );
    }
  }
}


// gets rid of the seq from the provided collections, moves it around the class-level
// collections as appropriate, deletes if that is ok.
void AssemblyJobNearPerfect::removeSeqHelper(ScoredSeqNormalized* oldSeq,
					     WindowStartAndLenCarrier* wslc,
					     map<string, set<ScoredSeqNormalized*>* >* windowToSubstrToSeqs){
  // remove the sequence from the front/back maps
  // make sure that the seq has been put in the table; if it has, remove it
  string oldSeqString = oldSeq->getSeq('+');

  for (int n = 0; n < wslc->_numWindows; ++n){
    string subSeq = oldSeqString.substr(wslc->_winStart[n],wslc->_winLen[n]);
    map<string,set<ScoredSeqNormalized*>* >::iterator matchIt = windowToSubstrToSeqs[n].find(subSeq);
    if ( matchIt != windowToSubstrToSeqs[n].end() ){
      matchIt->second->erase(oldSeq);
      if ( matchIt->second->size() == 0 ){
	delete matchIt->second;
	windowToSubstrToSeqs[n].erase( matchIt );
      }
    }
  }

  // delete the seq and its carrier
  ScoredSeq* oldInnerSeq = oldSeq->getNested();
  delete oldSeq;
  if (_inputSeqs.count(oldInnerSeq) == 0){ delete oldInnerSeq; }
}


// gets the matches satisfying the requirements as efficiently as possible
// using the front/back data structures and ungapped alignment technique
// does not delete anything, does not modify class data structures
// MODIFIES: legitMatches
void AssemblyJobNearPerfect::getLegitMatchesHelper(ScoredSeq* inputSeq, char sense,
						   WindowStartAndLenCarrier* wslc, long maxMisNum,
						   map<string, set<ScoredSeqNormalized*>* >* windowToSubstrToSeqs,
						   map<ScoredSeqNormalized*,Alignment*>* legitMatches){

  // establish the query sequences
  long seqLen = inputSeq->size();
  string keySeq = inputSeq->getSeq(sense);

  string* subSeqArray = new string[ wslc->_numWindows ];
  for (int n = 0; n < wslc->_numWindows; ++n){
    subSeqArray[n] = keySeq.substr(wslc->_winStart[n],wslc->_winLen[n]);
  }
  set<ScoredSeqNormalized*> allCandidateMatches;

  long allMatchSum = long( pow(double(2), double(wslc->_numWindows) ) ) - 1;
  for (long n = 0; n < wslc->_numWindows; ++n){
    map<string,set<ScoredSeqNormalized*>* >::iterator matchIt = windowToSubstrToSeqs[n].find( subSeqArray[n] );
    if ( matchIt != windowToSubstrToSeqs[n].end() ){
      for (set<ScoredSeqNormalized*>::iterator it = matchIt->second->begin(); it != matchIt->second->end(); ++it){
	(*it)->addCounts( long( pow(double(2), double(n)) ) );
	allCandidateMatches.insert( (*it) );
      }
    }
  }

  // check for legitimacy of candidate matches
  for (set<ScoredSeqNormalized*>::iterator candIt = allCandidateMatches.begin(); candIt != allCandidateMatches.end(); ++candIt){

    ScoredSeqNormalized* candSeq = (*candIt);
    bool legitMatchFound = false;

    // figure out if there is a legit match
    long candSeqCount = candSeq->getCount();
    if ( candSeqCount == allMatchSum ){ legitMatchFound = true; }
    else {
      // see if the number of windows with matches was sufficient for a quality match
      long numGoodWindows = 0;
      bool* isWindowGood = new bool[wslc->_numWindows];
      for (int n = 0; n < wslc->_numWindows; ++n){
	long binaryCount = ( candSeqCount % long( pow(double(2), double(n+1)) ) ) / long( pow(double(2), double(n) ) );
	numGoodWindows += binaryCount;
	isWindowGood[n] = bool(binaryCount);
      }
      if (wslc->_numWindows - numGoodWindows <= maxMisNum){
	// do the real comparison so that it will stop doing comparisons if the errors have over-accumulated
	legitMatchFound = true;
	long numRemainingMisses = maxMisNum;
	long n = 0;
	while (legitMatchFound and n < wslc->_numWindows){
	  // don't re-check perfect windows
	  if (! isWindowGood[n] ){
	    string compareSeq = candSeq->getSeq('+').substr(wslc->_winStart[n],wslc->_winLen[n]);
	    long numMis = compareSeqHelper(subSeqArray[n], compareSeq, sense, numRemainingMisses);
	    if (numMis > numRemainingMisses){ legitMatchFound = false; }
	    else { numRemainingMisses -= numMis; }
	  }
	  n++;
	}
      }
      delete [] isWindowGood;
    }

    // so that the opposite-strand matches can be found clearly
    candSeq->resetCount();

    // if there is a legit match, add the alignment to the collection of legit matches
    if (legitMatchFound){
      Alignment* legitMatch = new Alignment( inputSeq, candSeq, sense );
      legitMatch->addLinkageBlock(inputSeq,0,0,seqLen);
      legitMatch->lock();
      // check if there is already a legit match to the other strand
      map<ScoredSeqNormalized*,Alignment*>::iterator foundIt = legitMatches->find( candSeq );
      if ( foundIt == legitMatches->end() ){
	legitMatches->insert( pair<ScoredSeqNormalized*,Alignment*> (candSeq,legitMatch) );
      } else if ( foundIt->second->score( _matchCountAsm, false ) < legitMatch->score( _matchCountAsm, false ) ){
	delete foundIt->second; // delete the old alignment, replace with the better alignment
	foundIt->second = legitMatch;
      }
    }
  }

  delete [] subSeqArray;
}


long AssemblyJobNearPerfect::compareSeqHelper(string seqA, string seqB, char sense, long maxMisNum){
  if (seqA == seqB){ return 0; }
  else if (maxMisNum==0){ return 1; }
  else {
    long numMissed = 0;
    long pos = 0;
    while ( numMissed <= maxMisNum and pos < seqA.size() ){
      numMissed += _mismatchCountAsm->match( seqA[pos], seqB[pos] );
      ++pos;
    }
    return numMissed;
  }
}


#endif
