/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PARAMSCYCLEMANAGEMENT_CPP
#define PARAMSCYCLEMANAGEMENT_CPP

#include "ParamsCycleManagement.h"
#include "AssemblyException.h"
#include "ScoredSeqSubseq.h"
#include <iostream>
using namespace::std;



EcoFilter* ParamsCycleManagement::defaultCountEcoFilter(){ return new EcoFilterMinCount(1.5); }
long ParamsCycleManagement::defaultMaxContigLinkages(){ return 5; }

// constructors
ParamsCycleManagement::ParamsCycleManagement(){}

ParamsCycleManagement::ParamsCycleManagement(int totalCycles) :
  _totalCycles(totalCycles),
  _outputInterval(1)
{
  _maxContigLinkages = defaultMaxContigLinkages();
  setupFilters();
  setupTrimming();
}
ParamsCycleManagement::ParamsCycleManagement(int totalCycles, long maxContigLinkages) :
  _totalCycles(totalCycles),
  _maxContigLinkages(maxContigLinkages),
  _outputInterval(1)
{
  setupFilters();
  setupTrimming();
}



void ParamsCycleManagement::setupFilters(){
  _sizeFilter = new EcoFilterNull();
  _otherFilters = new EcoFilter*[1];
  _numOtherFilters = 0;
}

void ParamsCycleManagement::setupTrimming(){
  _cycleWillTrim = new bool[ _totalCycles ];
  _cycleToMinScore = new float[ _totalCycles ];
  _cycleToMinLength = new long[ _totalCycles ]; // -1 for no min length
  for (int n = 0; n < _totalCycles; ++n){
    _cycleWillTrim[n] = false;
    _cycleToMinScore[n] = 0; // just a meaningless filler value
    _cycleToMinLength[n] = 1; // provides info about whether a min len will be imposed
  }
  _initialWillTrim = false;
  _initialMinScore = 0;
  _initialMinLength = 1; // cannot have a 0-length contig
}


ParamsCycleManagement::~ParamsCycleManagement(){
  delete [] _cycleWillTrim;
  delete [] _cycleToMinScore;
  delete [] _cycleToMinLength;
  delete _sizeFilter;
  for (int n = 0; n < _numOtherFilters; ++n){ delete _otherFilters[n]; }
  delete [] _otherFilters;
}


ParamsCycleManagement::ParamsCycleManagement(ParamsCycleManagement* pcm) :
  _totalCycles(pcm->_totalCycles),
  _maxContigLinkages(pcm->_maxContigLinkages)
{
  _sizeFilter = pcm->_sizeFilter->copy();
  _numOtherFilters = pcm->_numOtherFilters;
  _otherFilters = new EcoFilter*[ _numOtherFilters + 1 ];
  for (int n = 0; n < _numOtherFilters; ++n){
    _otherFilters[n] = pcm->_otherFilters[n]->copy();
  }
  setupTrimming();
  for (int n  = 0; n < _totalCycles; ++n){
    _cycleWillTrim[n] = pcm->_cycleWillTrim[n];
    _cycleToMinScore[n] = pcm->_cycleToMinScore[n];
    _cycleToMinLength[n] = pcm->_cycleToMinLength[n];
  }
  _initialWillTrim = pcm->_initialWillTrim;
  _initialMinScore = pcm->_initialMinScore;
  _initialMinLength = pcm->_initialMinLength;
  _outputInterval = pcm->_outputInterval;
}
ParamsCycleManagement* ParamsCycleManagement::copy(){ return new ParamsCycleManagement(this); }



void ParamsCycleManagement::setSizeFilter(EcoFilter* sizeFilter){
  delete _sizeFilter;
  _sizeFilter = sizeFilter->copy();
}
EcoFilter* ParamsCycleManagement::getSizeFilter(){ return _sizeFilter->copy(); }

void ParamsCycleManagement::addEcoFilter(EcoFilter* ecoFilter){
  EcoFilter** filterHolder = new EcoFilter*[ _numOtherFilters + 2 ];
  for (int n = 0; n < _numOtherFilters; ++n){ filterHolder[n] = _otherFilters[n]; }
  filterHolder[ _numOtherFilters ] = ecoFilter->copy();
  delete [] _otherFilters;
  _otherFilters = filterHolder;
  _numOtherFilters++;
}
EcoFilter* ParamsCycleManagement::getFullEcoFilter(){
  _otherFilters[ _numOtherFilters ] = _sizeFilter;
  return new EcoFilterMulti(_otherFilters, _numOtherFilters + 1);
}




// get the values
int ParamsCycleManagement::totalCycles(){ return _totalCycles; }
long ParamsCycleManagement::maxContigLinkages(){ return _maxContigLinkages; }


void ParamsCycleManagement::addTrim(int numPriorCycles, float coverageLevel){
  if (numPriorCycles == 0){
    _initialWillTrim = true;
    _initialMinScore = coverageLevel;
  } else {
    _cycleWillTrim[numPriorCycles - 1] = true;
    _cycleToMinScore[numPriorCycles - 1] = coverageLevel;
  }
}
void ParamsCycleManagement::addTrim(int numPriorCycles, float coverageLevel, long minLength){
  if (numPriorCycles == 0){
    _initialWillTrim = true;
    _initialMinScore = coverageLevel;
    _initialMinLength = minLength;
  } else {
    _cycleWillTrim[numPriorCycles - 1] = true;
    _cycleToMinScore[numPriorCycles - 1] = coverageLevel;
    _cycleToMinLength[numPriorCycles - 1] = minLength;
  }
}


void ParamsCycleManagement::executeTrim(set<ScoredSeq*>* contigSet, int cycleNum){
  set<ScoredSeq*> discardSet;
  executeTrim(contigSet, cycleNum, &discardSet);
  for (set<ScoredSeq*>::iterator it = discardSet.begin(); it != discardSet.end(); ++it){
    (*it)->deepDelete();
  }
}
void ParamsCycleManagement::executeTrim(set<ScoredSeq*>* contigSet, int cycleNum, set<ScoredSeq*>* discardSet){
  bool willTrim;
  float minScore;
  long minLength;
  if( cycleNum == -1 ){
    willTrim = _initialWillTrim;
    minScore = _initialMinScore;
    minLength = _initialMinLength;
  } else {
    willTrim = _cycleWillTrim[cycleNum];
    minScore = _cycleToMinScore[cycleNum];
    minLength = _cycleToMinLength[cycleNum];
  }
  if (willTrim) {
    set<ScoredSeq*> kept;
    for (set<ScoredSeq*>::iterator it = contigSet->begin(); it != contigSet->end(); ++it){
      ScoredSeq* input = (*it);
      ScoredSeq* trimmed = trimThisContig( input, minScore, minLength );
      if (input == trimmed){
	kept.insert( input );
      } else {
	discardSet->insert( input );
	if ( trimmed != 0 ){ kept.insert( trimmed ); }
      }
    }
    contigSet->clear();
    contigSet->insert( kept.begin(), kept.end() );
  }
}

ScoredSeq* ParamsCycleManagement::trimThisContig(ScoredSeq* old, float minScore, long minLength){
  long contigSize = old->size();
  long startPos = 0;
  long endPos = contigSize - 1;
  while (startPos < endPos and old->scoreAtPosition(startPos,'+') < minScore){ startPos++; }
  while (endPos >= startPos and old->scoreAtPosition(endPos,'+') < minScore){ endPos--; }
  // a null will be returned if the resulting contig isn't long enough
  if (endPos - startPos + 1 < minLength){ return 0; }
  // no trim, so return the original
  else if (startPos == 0 and endPos == contigSize - 1){ return old; }
  else {
    ScoredSeq* tempSeq = new ScoredSeqSubseq(old, startPos, endPos - startPos + 1);
    ScoredSeq* returnSeq = tempSeq->shallowCopy();
    delete tempSeq;
    return returnSeq;
  }
}


void ParamsCycleManagement::setOutputInterval(int interval){
  if (interval < 1){ throw AssemblyException::ArgError("PCM output interval needs to be >= 1."); }
  _outputInterval = interval;
}
bool ParamsCycleManagement::outputThisCycle(int cycleNum){
  if (cycleNum < 1){ throw AssemblyException::ArgError("PCM::outputThisCycle, there is no cycle less than 1."); }
  return (cycleNum % _outputInterval == 0);
}

#endif
