/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef READPAIRFILTER_CPP
#define READPAIRFILTER_CPP

#include "ReadPairFilter.h"
#include "AssemblyException.h"
#include "Alignment.h"
#include "ScoredSeqCollectionBwt.h"
#include "MatchSeqTest.h"
#include "AssemblyJobPerfect.h"


ReadPairFilter::~ReadPairFilter(){}


ReadPairFilterNull::ReadPairFilterNull(){}
ReadPairFilterNull::~ReadPairFilterNull(){}
ReadPairFilterNull* ReadPairFilterNull::copy(){ return new ReadPairFilterNull(); }
bool ReadPairFilterNull::isPairOk(ScoredSeq* readA, ScoredSeq* readB){ return true; }


ReadPairFilterHomoPolymer::ReadPairFilterHomoPolymer(){}
ReadPairFilterHomoPolymer::ReadPairFilterHomoPolymer(long maxHomoPolymer) : _maxHomoPolymer(maxHomoPolymer){}
ReadPairFilterHomoPolymer::~ReadPairFilterHomoPolymer(){}
ReadPairFilterHomoPolymer* ReadPairFilterHomoPolymer::copy(){ return new ReadPairFilterHomoPolymer(_maxHomoPolymer); }
bool ReadPairFilterHomoPolymer::isPairOk(ScoredSeq* readA, ScoredSeq* readB){
  return (isReadOk(readA) and isReadOk(readB));
}
bool ReadPairFilterHomoPolymer::isReadOk(ScoredSeq* read){
  long rs = read->size();
  long pos = 0;
  while (pos < rs and read->nucAtPosition(pos,'+')=='N' and pos <= _maxHomoPolymer){ pos++; }
  // considers Ns to extend the prior nuc
  long currentCount = pos;
  // considers Ns to extend the following nuc
  long futureCount = pos;
  // the highest count so far (will be continuously updated)
  long maxCount = pos;
  // this is literally the prior nucleotide identity (at pos 0, there was no ID, hence N)
  char priorNuc = 'N';
  // N as the prior nuc is only valid as an initial condition for this variable
  char priorLegitNuc = 'N';

  while (pos < rs and maxCount <= _maxHomoPolymer){
    char currentNuc = read->nucAtPosition(pos,'+');
    if (currentNuc == 'N'){
      currentCount++;
      futureCount++;
    } else {
      if (currentNuc == priorLegitNuc){
	currentCount++;
	futureCount++;
      } else if (priorNuc == 'N'){
	currentCount = 1;
	futureCount++;
      } else {
	currentCount = 1;
	futureCount = 0;
      }
      priorLegitNuc = currentNuc;
    }
    priorNuc = currentNuc;
    pos++;
    if (currentCount > maxCount){ maxCount = currentCount; }
    if (futureCount > maxCount){ maxCount = futureCount; }
  }
  return (maxCount <= _maxHomoPolymer);
}





ReadPairFilterMulti::ReadPairFilterMulti(){}
ReadPairFilterMulti::ReadPairFilterMulti(ReadPairFilter** testArray, int numTests) : _numTests(numTests){
  _testArray = new ReadPairFilter*[ _numTests ];
  for (int n = 0; n < _numTests; ++n){ _testArray[n] = testArray[n]->copy(); }
}
ReadPairFilterMulti::~ReadPairFilterMulti(){
  for (int testNum = 0; testNum < _numTests; ++testNum){ delete _testArray[testNum]; }
  delete [] _testArray;
}
ReadPairFilterMulti* ReadPairFilterMulti::copy(){ return new ReadPairFilterMulti(_testArray,_numTests); }
bool ReadPairFilterMulti::isPairOk(ScoredSeq* readA, ScoredSeq* readB){
  bool testsPass = true;
  int testNum = 0;
  while (testsPass and testNum < _numTests){
    testsPass = _testArray[testNum]->isPairOk(readA, readB);
    testNum++;
  }
  return testsPass;
}


#endif

