/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef SCOREDSEQMONOSCORE_CPP
#define SCOREDSEQMONOSCORE_CPP

#include "ScoredSeqMonoScore.h"

#include "ScoredSeq.h"
#include "nucleicAcidUtilities.h"
#include "AssemblyException.h"
#include <cstring>
#include <omp.h>
using namespace::std;

ScoredSeqMonoScore::ScoredSeqMonoScore(){}; //default constructor

void ScoredSeqMonoScore::constructorSeqHelper(string seq){
  _size = seq.size();
  _seq = new char [_size+1];
  strcpy (_seq, seq.c_str());
  _rcSeq = nucleicAcidUtilities::reverseComplement(_seq,_size);
}

ScoredSeqMonoScore::ScoredSeqMonoScore(string seq, int score){
  constructorSeqHelper(seq);
  _score = float(score);
  _pe = 0;
  _thisIsAlive = true;
  //OK();
}
ScoredSeqMonoScore::ScoredSeqMonoScore(string seq, long score){
  constructorSeqHelper(seq);
  _score = float(score);
  _pe = 0;
  _thisIsAlive = true;
  //OK();
}
ScoredSeqMonoScore::ScoredSeqMonoScore(string seq, float score) :
  _score(score) {
  constructorSeqHelper(seq);
  _pe = 0;
  _thisIsAlive = true;
  //OK();
}

// CHAR ARRAYS

ScoredSeqMonoScore::ScoredSeqMonoScore(char* seq, int score, long size) :
  _size(size)
{
  _score = float(score);
  _seq = new char [_size+1];
  _seq[_size] = '\0';
  for (long n = 0; n < _size; ++n){ _seq[n] = seq[n]; }
  _rcSeq = nucleicAcidUtilities::reverseComplement(_seq,_size);
  _pe = 0;
  _thisIsAlive = true;
  //OK();
}
ScoredSeqMonoScore::ScoredSeqMonoScore(char* seq, long score, long size) :
  _size(size)
{
  _score = float(score);
  _seq = new char [_size+1];
  _seq[_size] = '\0';
  for (long n = 0; n < _size; ++n){ _seq[n] = seq[n]; }
  _rcSeq = nucleicAcidUtilities::reverseComplement(_seq,_size);
  _pe = 0;
  _thisIsAlive = true;
  //OK();
}
ScoredSeqMonoScore::ScoredSeqMonoScore(char* seq, float score, long size) :
  _size(size),
  _score(score) {
  _seq = new char[_size+1];
  _seq[_size] = '\0';
  for (long n = 0; n < _size; ++n){ _seq[n] = seq[n]; }
  _rcSeq = nucleicAcidUtilities::reverseComplement(_seq,_size);
  _pe = 0;
  _thisIsAlive = true;
  //OK();
}


ScoredSeqMonoScore::ScoredSeqMonoScore(const ScoredSeqMonoScore& r){
  throw AssemblyException::ImplementationError("not implemented");
}

ScoredSeqMonoScore::~ScoredSeqMonoScore(){
  bool deleteSelf = true; // keeps track if the reference fields need to be destroyed
  if (_thisIsAlive){
    makeDead();
    if ( _pe != 0 ){
      if ( _pe->isAlive() ) {
	// create a "copy" with the same references using function calls
	ScoredSeqMonoScore * copyOfThis = new ScoredSeqMonoScore( string(_seq), _score );
	copyOfThis->addPairedEnd( _pe );
	copyOfThis->makeDead();
	_pe->replaceDeadPe( copyOfThis ); // pass that copy to the paired end
	deleteSelf = false;
      } else {
	delete _pe;
	_pe = 0;
      }
    }
  }
  delete [] _seq;
  delete [] _rcSeq;
}

void ScoredSeqMonoScore::OK(){
}

float ScoredSeqMonoScore::scoreAtPosition(long position, char sense) {
  //OK();
  if ( nucAtPosition(position,sense) == 'N' ){ return 0; }
  else{ return _score; }
}

float ScoredSeqMonoScore::linkAfterPosition(long position, char sense){
  //OK();
  if ( position >= _size - 1 ){
    if ( position == _size - 1 ){
      throw AssemblyException::LengthError("last position does not have a linkage score AFTER it");
    } else { throw AssemblyException::LengthError("position is outside of the scope of the sequence"); }
  }
  return _score;
}


char ScoredSeqMonoScore::nucAtPosition(long position, char sense) {
  //OK();
  //PROBLEMS: define "length", define a more-efficient method for
  //dealing with sense. define "reverseCompChar" method.
  if (position < 0 or position >= _size){
    throw AssemblyException::LengthError("position is outside of the scope of the sequence"); }
  if (sense=='+'){
    return _seq[position];
  } else if (sense=='-'){
    return _rcSeq[position];
    //return nucleicAcidUtilities::reverseCompChar(_seq[_seq.size() - position - 1]);
  } else {
    throw AssemblyException::ArgError("sense must be '+' or '-'");
  }
}

string ScoredSeqMonoScore::getSeq(char sense) {
  //OK();
  if (sense=='+'){
    return string(_seq);
  } else {
    return string(_rcSeq);
  }
}
string ScoredSeqMonoScore::getSubseq(long position, long length, char sense){
  //OK();
  if (sense=='+'){
    return string(_seq).substr(position,length);
  } else {
    return string(_rcSeq).substr(position,length);
  }
}

long ScoredSeqMonoScore::size() {
  //OK();
  return _size;
}

void ScoredSeqMonoScore::buffer(){}
void ScoredSeqMonoScore::unbuffer(){}
void ScoredSeqMonoScore::bottomBuffer(){}
void ScoredSeqMonoScore::deepDelete(){ delete this; }
void ScoredSeqMonoScore::deepUnbuffer(){} // unbuffer doesn't do anything, so no need to call it


bool ScoredSeqMonoScore::hasPairedEnd() { return _pe != 0; }
ScoredSeq * ScoredSeqMonoScore::getPairedEnd() {
  if (! hasPairedEnd()) {
    throw AssemblyException::LogicError("This ScoredSeq doesn't have a paired end.");
  }
  _pe->makeAlive();
  return _pe;
}
ScoredSeq * ScoredSeqMonoScore::getTempPairedEnd() {
  if (! hasPairedEnd()) {
    throw AssemblyException::LogicError("This ScoredSeq doesn't have a paired end.");
  }
  return _pe;
}

void ScoredSeqMonoScore::makeDead(){ _thisIsAlive = false; }
void ScoredSeqMonoScore::makeAlive(){ _thisIsAlive = true; }
bool ScoredSeqMonoScore::isAlive(){ return _thisIsAlive; }

void ScoredSeqMonoScore::replaceDeadPe(ScoredSeqMonoScore * deadReplacement){
  if (_pe == deadReplacement) {
    throw AssemblyException::ArgError("PE original and copy have same memory address.");
  }
  deadReplacement->makeDead();
  _pe = deadReplacement;
}


void ScoredSeqMonoScore::addPairedEnd(ScoredSeqMonoScore * ss){
  _pe = ss;
}

ScoredSeq * ScoredSeqMonoScore::shallowCopy(){
  //OK();
  return new ScoredSeqMonoScore(string(_seq), _score);
}
ScoredSeq * ScoredSeqMonoScore::flipCopy(){
  //OK();
  return new ScoredSeqMonoScore(string(_rcSeq), _score);
}


bool ScoredSeqMonoScore::isNested(){ return false; }



#endif
